use wasm_bindgen::prelude::*;
use web_sys::console;
use ironecho;
use std::fmt;
use std::f64::consts::PI;
use num;
use num::Complex;

// When the `wee_alloc` feature is enabled, this uses `wee_alloc` as the global
// allocator.
//
// If you don't want to use `wee_alloc`, you can safely delete this.
#[cfg(feature = "wee_alloc")]
#[global_allocator]
static ALLOC: wee_alloc::WeeAlloc = wee_alloc::WeeAlloc::INIT;


// This is like the `main` function, except for JavaScript.
#[wasm_bindgen(start)]
pub fn main_js() -> Result<(), JsValue> {
    // This provides better error messages in debug mode.
    // It's disabled in release mode so it doesn't bloat up the file size.
    #[cfg(debug_assertions)]
    console_error_panic_hook::set_once();


    // Your code goes here!
    console::log_1(&JsValue::from_str("Hello world!"));

    Ok(())
}

#[wasm_bindgen]
#[derive(Copy,Clone)]
pub struct SimpleLayer {
    pub vp: f32,
    pub vs: f32,
    pub thick: f32,
    pub rho: f32,
}
impl fmt::Display for SimpleLayer {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "{} {} {} {} 1000.0 500.0 10000.0 0.0 10000.0 0.0", self.thick, self.vp, self.vs, self.rho)
    }
}

#[wasm_bindgen]
#[derive(Copy,Clone)]
pub struct WMomentTensor {
    pub nn: f32,
    pub ne: f32,
    pub nd: f32,
    pub ee: f32,
    pub ed: f32,
    pub dd: f32,
}
impl fmt::Display for WMomentTensor {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "{} {} {} {} {} {}", self.nn, self.ne, self.nd, self.ee, self.ed, self.dd)
    }
}
#[wasm_bindgen]
#[derive(Copy,Clone)]
pub struct SimpleParams {
    pub pmax: f32,
    pub fmin: f32,
    pub fmax: f32,
    pub num_points: usize,
    pub dist: f32,
    pub az: f32,
    pub source_depth: f32,
    pub moment_tensor: WMomentTensor
}
impl fmt::Display for SimpleParams {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "0.0 0.0 {} {} 1.0", self.pmax, self.pmax*1.05);
        writeln!(f, "{} {} {} {}", self.fmin, self.fmax, self.fmax, self.num_points);
        writeln!(f, "{} {}", self.dist, self.az);
        writeln!(f, "1");
        writeln!(f, "{}", self.source_depth);
        writeln!(f, "0.0");
        write!(f, "{}", self.moment_tensor)
    }
}
#[wasm_bindgen]
impl SimpleParams {

    pub fn to_string(&self) -> String {
        format!("{}",self)
    }
}

#[wasm_bindgen]
pub struct SimpleModel {
    layers: Vec<SimpleLayer>,
    pub params: SimpleParams,
}

impl fmt::Display for SimpleModel {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        writeln!(f, "{}",self.layers.len());
        for l in &self.layers {
            write!(f, "{}", l);
        }
        write!(f, "{}", self.params)
    }
}
#[wasm_bindgen]
impl SimpleModel {
    pub fn get_vp(&self, layer_num: usize) -> Result<f32, JsValue> {
        if layer_num >= self.layers.len() {
            return Err(JsValue::from_str("not enough layers"));
        }
        let vp = self.layers[layer_num].vp;
        Ok(vp)
    }
    pub fn get_layer(&self, i: usize) -> Result<SimpleLayer, JsValue> {
        if i < self.layers.len() {
            let l = self.layers[i];
            return Ok(l);
        }
        Err(JsValue::from_str("not enough layers"))
    }
    pub fn to_string(&self) -> String {
        format!("{}",self)
    }
    pub fn setNumPoints(&mut self, val: usize) {
        self.params.num_points = val;
    }
    pub fn setDistance(&mut self, val: f32) {
        self.params.dist = val;
    }
}


#[wasm_bindgen]
pub fn gen_model() -> Result<SimpleModel, JsValue> {

    let lines = "3
35.00 6.50 3.50 2.70 1000.00 500.00 10000.00 0.00 10000.00 0.00
30.00 8.10 4.67 3.32 1000.00 500.00 10000.00 0.00 10000.00 0.00
0.00 8.10 4.67 3.32 1000.00 500.00 10000.00 0.00 10000.00 0.00
0.0 0.0 0.3 0.4 1.0
0.0 1.0 1.0 1024
200 45
1
10.001
0
0.0 0.707 -0.707 0.0 0.0 0.0
";
    let simpModel = gen_model_from_string(lines).unwrap();
    Ok(simpModel)
}


#[wasm_bindgen]
pub fn gen_model_from_string(lines: &str) -> Result<SimpleModel, JsValue> {
    let lines: Vec<&str> = lines.lines().collect::<Vec<&str>>();
    let (ger_model, ger_params) = ironecho::modelfile::load_from_lines(&lines);
    let mut layers = Vec::<SimpleLayer>::new();
    for i in 0..ger_model.al0.len() {
        layers.push(SimpleLayer {
            vp: ger_model.al0[i] as f32,
            vs: ger_model.be0[i] as f32,
            thick: ger_model.thik[i] as f32,
            rho: ger_model.rho[i] as f32,
         });
    }
    let mt = &ger_params.moment_tensor;
    let params = SimpleParams {
        pmax: ger_params.slowness.highpass  as f32,
        fmin: ger_params.frequency.min as f32,
        fmax: ger_params.frequency.max as f32,
        source_depth: ger_params.source_depth_list[0] as f32,
        num_points: ger_params.frequency.numtimepoints,
        dist: ger_params.distance.as_vec()[0] as f32,
        az: ger_params.azimuth as f32,
        moment_tensor: WMomentTensor {
            nn: mt.nn as f32,
            ne: mt.ne as f32,
            nd: mt.nd as f32,
            ee: mt.ee as f32,
            ed: mt.ed as f32,
            dd: mt.dd as f32
        },
    };
    let simpModel = SimpleModel {
        layers,
        params
    };
    Ok(simpModel)
}


#[wasm_bindgen]
pub struct FreqResult {
    specResult: ironecho::calc::SpecResult,
    model: ironecho::kennet::Model,
    params: ironecho::modelfile::RunParameters,
}
#[wasm_bindgen]
impl FreqResult {
    pub fn freq_len(&self) -> usize {
        self.params.frequency.numtimepoints
    }
    pub fn z(&self, reduce: f64, offset: f64) -> Vec<f64> {
        let (ifmin, ifmax) = self.params.frequency.get_ifmin_ifmax();
        console::log_1(&JsValue::from_str(&format!("reduce: {} offset: {}", reduce, offset)));
        console::log_1(&JsValue::from_str(&format!("npts: {} ifmin: {} ifmax: {}  spec: {}",self.params.frequency.numtimepoints, ifmin, ifmax, self.specResult.u0t.len())));
        let mut out = vec![num::Complex::new(0.0,0.0); self.params.frequency.numtimepoints/2+1];
        for i in ifmin..=ifmax {
            out[i] = self.specResult.u0t[i-ifmin][0][0];
        }
        self.as_timeseries(out, reduce, offset)
    }
    pub fn r(&self, reduce: f64, offset: f64) -> Vec<f64> {
        let mut out = vec![Complex::new(0.0,0.0); self.params.frequency.numtimepoints/2+1];
        let (ifmin, ifmax) = self.params.frequency.get_ifmin_ifmax();
        for i in ifmin..=ifmax {
            out[i] = self.specResult.w0t[i-ifmin][0][0];
        }
        self.as_timeseries(out, reduce, offset)
    }
    pub fn t(&self, reduce: f64, offset: f64) -> Vec<f64> {
        let mut out = vec![Complex::new(0.0,0.0); self.params.frequency.numtimepoints/2+1];
        let (ifmin, ifmax) = self.params.frequency.get_ifmin_ifmax();
        for i in ifmin..=ifmax {
            out[i] = self.specResult.tnt[i-ifmin][0][0];
        }
        self.as_timeseries(out, reduce, offset)
    }
    pub fn time_reduce(&self, reduceVel: f64, offset: f64) -> f64 {
        self.params.distance.as_vec()[0] / reduceVel + offset
    }
    fn as_timeseries(&self, x: Vec<Complex<f64>>, reduceVel: f64, offset: f64) -> Vec<f64> {
        let delf = self.params.frequency.get_delf();
        let timeReduce = self.time_reduce(reduceVel, offset);
        let reduceShift = timeReduce * 2.0 * PI*delf;
        let (ifmin, ifmax) = self.params.frequency.get_ifmin_ifmax();
        let nffpts = ifmax - ifmin +1;
        let mut x = x;
        for i in ifmin..=ifmax {
            let c = ironecho::util::cmplx(0.0, i as f64 *delf*2.0*PI);
            let sh = ironecho::util::cmplx(0., (i as f64)*reduceShift).exp();
            x[i] = x[i] * c * sh;
        }
        let mut out = vec![0.0; self.params.frequency.numtimepoints];
        let N = self.params.frequency.numtimepoints;
        out[0] = x[0].re; // zero freq is real
        for i in 1..(N/2) {
            out[i] = x[i].re;
            out[N-i] = x[i].im;
        }
        out[N/2] = x[N/2].re; // ny freq real
        out
    }
}


#[wasm_bindgen]
pub fn calc_it(model: &SimpleModel) -> Result<FreqResult, JsValue> {
    let lines = model.to_string();
    let lines: Vec<&str> = lines.lines().collect::<Vec<&str>>();
    let (ger_model, ger_params) = ironecho::modelfile::load_from_lines(&lines);
    let mut specResult = ironecho::calc::calc(&ger_model, &ger_params);

    let delf = &ger_params.frequency.nyquist * 2.0 / ger_params.frequency.numtimepoints as f64;
    let N = specResult.u0t.len();
    let L = 2*(N-1);
    let reduceVel = 8.1;
    let offset = -10.0;
    let timeReduce = ger_params.distance.as_vec()[0] / reduceVel + offset;
    let reduceShift = timeReduce * 2.0 * PI*delf;
    for i in 0..N {
        let c = ironecho::util::cmplx(0.0, i as f64 *delf*2.0*PI);
        let sh = ironecho::util::cmplx(0., (i as f64)*reduceShift).exp();
        specResult.u0t[i][0][0] = specResult.u0t[i][0][0] * c * sh;
        specResult.w0t[i][0][0] = specResult.w0t[i][0][0] * c * sh;
        specResult.tnt[i][0][0] = specResult.tnt[i][0][0] * c * sh;
    }
    let mut outZ = vec![0.0; L];
    let mut outR = vec![0.0; L];
    let mut outT = vec![0.0; L];
    outZ[0] = specResult.u0t[0][0][0].re; // zero freq real
    outR[0] = specResult.w0t[0][0][0].re; // zero freq real
    outT[0] = specResult.tnt[0][0][0].re; // zero freq real
    for i in 1..(N-1) {
        let c = &specResult.u0t[i][0][0];
        outZ[i] = c.re;
        outZ[L-i] = c.im;
        let c = &specResult.w0t[i][0][0];
        outR[i] = c.re;
        outR[L-i] = c.im;
        let c = &specResult.tnt[i][0][0];
        outT[i] = c.re;
        outT[L-i] = c.im;
    }
    outZ[N] = specResult.u0t[N-1][0][0].re; // ny freq real
    outR[N] = specResult.w0t[N-1][0][0].re; // ny freq real
    outT[N] = specResult.tnt[N-1][0][0].re; // ny freq real

    Ok(FreqResult {specResult, model: ger_model, params: ger_params})
}
